@@ -1,12 +1,14 @@
 """machinable"""
 __all__ = [
     "Element",
+    "Interface",
     "Execution",
-    "Experiment",
+    "Component",
     "Project",
-    "Record",
     "Storage",
     "Mixin",
+    "Index",
+    "Scope",
     "mixin",
     "Schedule",
     "get",
@@ -22,25 +24,26 @@ else:
     import importlib_metadata
 
 from machinable.cli import from_cli
+from machinable.component import Component
 from machinable.element import Element
 from machinable.execution import Execution
-from machinable.experiment import Experiment
+from machinable.index import Index
+from machinable.interface import Interface
 from machinable.mixin import Mixin, mixin
 from machinable.project import Project
-from machinable.record import Record
 from machinable.schedule import Schedule
+from machinable.scope import Scope
 from machinable.settings import get_settings
 from machinable.storage import Storage
-from machinable.types import Optional, VersionType
+from machinable.types import Optional, Union, VersionType
 
 
 def get(
-    module: Optional[str] = None,
+    module: Union[str, Element, None] = None,
     version: VersionType = None,
-    predicate: Optional[str] = get_settings().default_predicate,
     **kwargs,
-) -> Element:
-    return Element.get(module, version, predicate, **kwargs)
+) -> Interface:
+    return Interface.get(module, version, **kwargs)
 
 
 def get_version() -> str:
