@@ -9,7 +9,6 @@ from functools import reduce
 from json import dumps
 from pprint import pprint
 
-from machinable.settings import get_settings
 from machinable.types import VersionType
 
 long = int
@@ -18,7 +17,7 @@ basestring = str
 
 
 if TYPE_CHECKING:
-    from machinable.experiment import Experiment
+    from machinable.component import Component
 
 
 def _get_value(val):
@@ -1380,42 +1379,31 @@ class ElementCollection(Collection):
     def filter_by_module(self, module):
         return self.filter(lambda x: x.module == module)
 
-    def filter_by_predicate(
+    def filter_by_context(
         self,
         module: str,
         version: VersionType = None,
-        predicate: str = get_settings().default_predicate,
         **kwargs,
     ):
         from machinable import Element
 
-        try:
-            instance = Element.make(module, version, **kwargs)
-        except ModuleNotFoundError:
-            from machinable.element import equalversion
+        instance = Element.make(module, version, **kwargs)
 
-            # use direct comparison fallback
-            return self.filter(
-                lambda x: bool(
-                    x.module == module and equalversion(x.version(), version)
-                )
-            )
-
-        return self.filter(lambda x: x.matches(instance, predicate))
+        return self.filter(lambda x: x.matches(instance))
 
     def singleton(
         self,
         module: str,
         version: VersionType = None,
-        predicate: str = get_settings().default_predicate,
         **kwargs,
-    ) -> Union[Any, "Experiment"]:
+    ) -> Union[Any, "Component"]:
         from machinable import Element
 
         instance = Element.make(module, version, **kwargs)
+        context = instance.compute_context()
 
         for candidate in self:
-            if candidate.matches(instance, predicate):
+            if candidate.matches(instance, context):
                 return candidate
 
         return instance
@@ -1424,7 +1412,11 @@ class ElementCollection(Collection):
         return f"Elements <{len(self.items)}>"
 
 
-class ExperimentCollection(ElementCollection):
+class InterfaceCollection(ElementCollection):
+    pass
+
+
+class ComponentCollection(InterfaceCollection):
     def __str__(self):
         if len(self.items) > 15:
             items = ", ".join([repr(item) for item in self.items[:5]])
@@ -1432,29 +1424,20 @@ class ExperimentCollection(ElementCollection):
             items += ", ".join([repr(item) for item in self.items[-5:]])
         else:
             items = ", ".join([repr(item) for item in self.items])
-        return f"Experiments ({len(self.items)}) <{items}>"
+        return f"Components ({len(self.items)}) <{items}>"
 
-    def launch(self) -> "ExperimentCollection":
-        """Executes all experiments in the collection"""
-        for experiment in self:
-            experiment.launch()
+    def launch(self) -> "ComponentCollection":
+        """Executes all components in the collection"""
+        for component in self:
+            component.launch()
 
         return self
 
-    def finished(self) -> "ExperimentCollection":
-        return self.filter(lambda x: x.is_finished())
-
-    def started(self) -> "ExperimentCollection":
-        return self.filter(lambda x: x.is_started())
-
-    def active(self) -> "ExperimentCollection":
-        return self.filter(lambda x: x.is_active())
-
-    def incomplete(self) -> "ExperimentCollection":
-        return self.filter(lambda x: x.is_incomplete())
-
 
 class ExecutionCollection(ElementCollection):
+    def __str__(self):
+        return f"Executions <{len(self.items)}>"
+
     def status(self, status="started"):
         """Filters the collection by a status attribute
 
@@ -1465,22 +1448,3 @@ class ExecutionCollection(ElementCollection):
             return self.filter(lambda item: getattr(item, "is_" + status)())
         except AttributeError as _ex:
             raise ValueError(f"Invalid status field: {status}") from _ex
-
-    def __str__(self):
-        return f"Executions <{len(self.items)}>"
-
-
-class RecordCollection(ElementCollection):
-    def as_dataframe(self):
-        """Returns collection as Pandas dataframe"""
-        import pandas
-
-        data = (
-            {k: [row[k] for row in self._items] for k in self._items[0].keys()}
-            if len(self._items) > 0
-            else {}
-        )
-        return pandas.DataFrame.from_dict(data)
-
-    def __str__(self):
-        return f"Records <{len(self.items)}>"
