@@ -1,101 +1,34 @@
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    Dict,
-    List,
-    Optional,
-    Tuple,
-    Union,
-)
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
 
 import collections
 import copy
 import json
-import os
-import shlex
-import stat
-from functools import wraps
+import sys
+
+if sys.version_info >= (3, 11):
+    from typing import Self
+else:
+    from typing_extensions import Self
 
 import arrow
+import dill as pickle
 import machinable
 import omegaconf
-import pydantic
-from flatten_dict import flatten
 from machinable import schema
-from machinable.collection import Collection
+from machinable.collection import ElementCollection
 from machinable.config import from_element, match_method, rewrite_config_methods
 from machinable.errors import ConfigurationError, MachinableError
 from machinable.mixin import Mixin
-from machinable.settings import get_settings
-from machinable.types import ElementType, VersionType
-from machinable.utils import Jsonable, unflatten_dict, update_dict
+from machinable.types import DatetimeType, ElementType, VersionType
+from machinable.utils import (
+    Jsonable,
+    sentinel,
+    serialize,
+    unflatten_dict,
+    update_dict,
+)
 from omegaconf import DictConfig, OmegaConf
 
-if TYPE_CHECKING:
-    from machinable.storage import Storage
-
-
-def belongs_to(f: Callable) -> Any:
-    @property
-    @wraps(f)
-    def _wrapper(self: "Element"):
-        name = f.__name__
-        if self.__related__.get(name, None) is None and self.is_mounted():
-            related_class = f()
-            use_cache = True
-            if isinstance(related_class, tuple):
-                related_class, use_cache = related_class
-            related = self.__model__._storage_instance.retrieve_related(
-                self.__model__._storage_id,
-                f"{self.kind.lower()}.{name}",
-            )
-            if related is None:
-                return None
-            element = related_class.from_model(related)
-            if not use_cache:
-                return element
-            self.__related__[name] = element
-
-        return self.__related__.get(name, None)
-
-    return _wrapper
-
-
-has_one = belongs_to
-
-
-def has_many(f: Callable) -> Any:
-    @property
-    @wraps(f)
-    def _wrapper(self: "Element") -> Any:
-        name = f.__name__
-        if self.__related__.get(name, None) is None and self.is_mounted():
-            args = f()
-            use_cache = True
-            if len(args) == 2:
-                related_class, collection = args
-            elif len(args) == 3:
-                related_class, collection, use_cache = args
-            else:
-                assert False, "Invalid number of relation arguments"
-            related = self.__model__._storage_instance.retrieve_related(
-                self.__model__._storage_id,
-                f"{self.kind.lower()}.{name}",
-            )
-            if related is None:
-                return None
-            collected = collection(
-                [related_class.from_model(r) for r in related]
-            )
-            if not use_cache:
-                return collected
-            self.__related__[name] = collected
-
-        return self.__related__.get(name, None)
-
-    return _wrapper
-
 
 class ConfigMethod:
     def __init__(self, element: "Element", prefix="config") -> None:
@@ -188,6 +121,12 @@ def defaultversion(
     if module is not None:
         return module, normversion(version)
 
+    # handle in-session defaults
+    if isinstance(element.default, (list, tuple)) and not isinstance(
+        element.default[0], str
+    ):
+        return element.default[0], element.default[1:] + normversion(version)
+
     default_version = normversion(element.default)
 
     if len(default_version) == 0:
@@ -215,9 +154,12 @@ def extract(
     if isinstance(compact_element, str):
         return compact_element, None
 
+    if isinstance(compact_element, omegaconf.listconfig.ListConfig):
+        compact_element = list(compact_element)
+
     if not isinstance(compact_element, (list, tuple)):
         raise ValueError(
-            f"Invalid component defintion. Expected list or str but found {compact_element}"
+            f"Invalid component defintion. Expected list or str but found {type(compact_element)}: {compact_element}"
         )
 
     if len(compact_element) == 0:
@@ -236,29 +178,33 @@ def extract(
     return compact_element[0], normversion(compact_element[1:])
 
 
-def equalversion(a: VersionType, b: VersionType) -> bool:
-    return json.dumps(normversion(a), sort_keys=True) == json.dumps(
-        normversion(b), sort_keys=True
+def equaljson(a: Any, b: Any) -> bool:
+    return json.dumps(a, sort_keys=True, default=serialize) == json.dumps(
+        b, sort_keys=True, default=serialize
     )
 
 
-def _filter_enderscores(m: Any):
+def equalversion(a: VersionType, b: VersionType) -> bool:
+    return equaljson(normversion(a), normversion(b))
+
+
+def filter_enderscores(m: Any):
     if not isinstance(m, collections.abc.Mapping):
         return m
     return {
-        k: _filter_enderscores(v) for k, v in m.items() if not k.endswith("_")
+        k: filter_enderscores(v) for k, v in m.items() if not k.endswith("_")
     }
 
 
-def _idversion_filter(value: Union[str, dict]) -> Union[str, dict, None]:
+def idversion_filter(value: Union[str, dict]) -> Union[str, dict, None]:
     if isinstance(value, str) and value.endswith("_"):
         return None
 
-    return _filter_enderscores(value)
+    return filter_enderscores(value)
 
 
 def idversion(version: VersionType) -> VersionType:
-    return normversion([_idversion_filter(v) for v in normversion(version)])
+    return normversion([idversion_filter(v) for v in normversion(version)])
 
 
 def transfer_to(src: "Element", destination: "Element") -> "Element":
@@ -267,25 +213,15 @@ def transfer_to(src: "Element", destination: "Element") -> "Element":
     return destination
 
 
-def resolve_custom_predicate(predicate: str, element: "Element"):
-    from machinable.project import Project
+def uuid_to_id(uuid: str) -> str:
+    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
 
-    custom = element.on_compute_predicate() or {}
-    if isinstance(element, Project):
-        custom.update(element.global_predicate() or {})
-    else:
-        custom.update(Project.get().provider().global_predicate() or {})
+    # convert uuid hex to base 62 of length 6
+    result = ""
+    for i in range(0, 6 * 2, 2):
+        result += alphabet[int(uuid[i : i + 2], 16) % 62]
 
-    if custom:
-        predicate = predicate.replace(
-            "*",
-            ",".join(
-                [k.replace("*", "") for k in custom.keys() if k.endswith("*")]
-            ),
-        )
-    return [
-        p.strip() for p in predicate.split(",") if p.strip() not in ("*", "")
-    ]
+    return result
 
 
 def instantiate(
@@ -316,56 +252,106 @@ class Element(Mixin, Jsonable):
         super().__init__()
         if Element._module_ is None:
             Element._module_ = self.__module__
-        self.__model__ = schema.Element(
+        self.__model__ = getattr(schema, self.kind, schema.Element)(
+            kind=self.kind,
             module=Element._module_,
             version=normversion(version),
             lineage=get_lineage(self),
+            _dump=pickle.dumps(self.__class__)
+            if Element._module_.startswith("__session__")
+            else None,
         )
-        self.__related__ = {}
         self.__mixin__ = None
         self.__mixins__ = {}
         self._config: Optional[DictConfig] = None
         self._predicate: Optional[DictConfig] = None
         self._cache = {}
-        self._deferred_data = {}
+        self._kwargs = {}
 
         Element._module_ = None
 
+    @property
+    def _enderscore_config(self):
+        if "enderscore_config" not in self._cache:
+            self._cache["enderscore_config"] = filter_enderscores(self.config)
+        return self._cache["enderscore_config"]
+
+    @property
+    def uuid(self) -> str:
+        return self.__model__.uuid
+
+    @property
+    def id(self) -> str:
+        return self.uuid[:6]
+
+    @property
+    def timestamp(self) -> float:
+        return self.__model__.timestamp
+
+    def created_at(self) -> DatetimeType:
+        return arrow.get(self.__model__.timestamp)
+
+    def version(
+        self, version: VersionType = sentinel, overwrite: bool = False
+    ) -> List[Union[str, dict]]:
+        if version is sentinel:
+            return self.__model__.version
+
+        if hasattr(self, "is_mounted") and self.is_mounted():
+            raise MachinableError(
+                f"Cannot change version of mounted element {self}"
+            )
+
+        if overwrite:
+            self.__model__.version = normversion(version)
+        else:
+            self.__model__.version.extend(normversion(version))
+
+        self._clear_caches()
+
+        return self.__model__.version
+
     @classmethod
     def set_default(
         cls,
-        module: Optional[str] = None,
+        module: Union[str, "Element", None] = None,
         version: VersionType = None,
     ) -> None:
+        if module is not None and not isinstance(module, str):
+            cls.default = [module] + normversion(version)
+            return
         cls.default = compact(module, version)
 
-    def as_default(self) -> "Element":
+    def as_default(self) -> Self:
         cls = getattr(machinable, self.kind, Element)
         cls.set_default(self.__model__.module, self.__model__.version)
 
         return self
 
+    @classmethod
+    def connected(cls) -> List["Element"]:
+        return _CONNECTIONS[cls.kind]
+
     @classmethod
     def get(
         cls,
-        module: Optional[str] = None,
+        module: Union[str, "Element", None] = None,
         version: VersionType = None,
-        predicate: Optional[str] = get_settings().default_predicate,
         **kwargs,
     ) -> "Element":
         if module is None and version is None:
             if len(_CONNECTIONS[cls.kind]) > 0:
                 return _CONNECTIONS[cls.kind][-1]
 
-        if module is None or predicate is None:
-            return cls.instance(module, version, **kwargs)
+        if module is None:
+            return cls.make(module, version, **kwargs)
 
-        return cls.singleton(module, version, predicate, **kwargs)
+        return cls.singleton(module, version, **kwargs)
 
     @classmethod
     def make(
         cls,
-        module: Optional[str] = None,
+        module: Union[None, str, "Element"] = None,
         version: VersionType = None,
         base_class: "Element" = None,
         **kwargs,
@@ -392,36 +378,35 @@ class Element(Mixin, Jsonable):
     @classmethod
     def instance(
         cls,
-        module: Optional[str] = None,
+        module: Union[None, str, "Element"] = None,
         version: VersionType = None,
         **kwargs,
     ) -> "Element":
         module, version = defaultversion(module, version, cls)
         return cls.make(module, version, base_class=cls, **kwargs)
 
-    def mount(self, storage: "Storage", storage_id: Any) -> bool:
-        if self.__model__ is None:
-            return False
-
-        self.__model__._storage_instance = storage
-        self.__model__._storage_id = storage_id
-
-        return True
-
-    def is_mounted(self) -> bool:
-        if self.__model__ is None:
-            return False
-
-        return (
-            self.__model__._storage_instance is not None
-            and self.__model__._storage_id is not None
-        )
+    @classmethod
+    def singleton(
+        cls,
+        module: Union[str, "Element"],
+        version: VersionType = None,
+        **kwargs,
+    ) -> "Element":
+        # no-op as elements do not have a storage representation
+        return cls.make(module, version, **kwargs)
 
     @classmethod
     def from_model(cls, model: schema.Element) -> "Element":
         if cls.__module__ != model.module:
             # re-instantiate element class
-            instance = cls.make(model.module)
+            if model.module and model.module.startswith("__session__"):
+                if model._dump is None:
+                    raise RuntimeError(
+                        f"Unable to restore element {model.module}"
+                    )
+                instance = cls.make(pickle.loads(model._dump))
+            else:
+                instance = cls.make(model.module)
         else:
             instance = cls()
 
@@ -429,97 +414,42 @@ class Element(Mixin, Jsonable):
 
         return instance
 
-    @classmethod
-    def find(cls, element_id: str, *args, **kwargs) -> Optional["Element"]:
-        from machinable.storage import Storage
+    def compute_context(self) -> Optional[Dict]:
+        """Computes the context contraints of the element
 
-        storage = Storage.get()
+        Returns:
+            Optional[Dict]: Maps constraint fields their required JSON-able values.
+                An empty dict means no constraits, i.e. all elements will be matched.
+                Returning None means full constraits, i.e. no element will be matched.
+        """
+        return {
+            "module": self.module,
+            "config": self._enderscore_config,
+            "predicate": self.compute_predicate(),
+        }
 
-        storage_id = getattr(storage, f"find_{cls.kind.lower()}")(
-            element_id, *args, **kwargs
-        )
+    def on_compute_predicate(self) -> Dict:
+        """Event to compute additional predicates that identify this element.
 
-        if storage_id is None:
-            return None
+        Returns:
+            Dict: Maps the names used during lookup to their JSON-able value.
+        """
+        return {}
 
-        return cls.from_storage(storage_id, storage)
+    def compute_predicate(self) -> Dict:
+        predicate = self.on_compute_predicate() or {}
 
-    @classmethod
-    def find_many(cls, elements: List[str]) -> "Collection":
-        return cls.collect([cls.find(element_id) for element_id in elements])
+        # apply scopes
+        for scope in _CONNECTIONS["Scope"]:
+            predicate.update(scope())
 
-    @classmethod
-    def find_by_predicate(
-        cls,
-        module: str,
-        version: VersionType = None,
-        predicate: Optional[str] = get_settings().default_predicate,
-        **kwargs,
-    ) -> "Collection":
-        from machinable.storage import Storage
-
-        storage = Storage.get()
-        try:
-            candidate = cls.make(module, version, **kwargs)
-        except ModuleNotFoundError:
-            return cls.collect([])
-
-        element_type = candidate.kind.lower()
-        handler = f"find_{element_type}_by_predicate"
-
-        if hasattr(storage, handler):
-            if predicate:
-                predicate = OmegaConf.to_container(
-                    OmegaConf.create(
-                        {
-                            p: candidate.predicate[p]
-                            for p in resolve_custom_predicate(
-                                predicate, candidate
-                            )
-                        }
-                    )
-                )
-            storage_ids = getattr(storage, handler)(module, predicate)
-        else:
-            storage_ids = []
-
-        return cls.collect(
-            [
-                cls.from_model(
-                    getattr(storage, f"retrieve_{element_type}")(storage_id)
-                )
-                for storage_id in storage_ids
-            ]
-        )
-
-    @classmethod
-    def singleton(
-        cls,
-        module: str,
-        version: VersionType = None,
-        predicate: Optional[str] = get_settings().default_predicate,
-        **kwargs,
-    ) -> "Collection[Element]":
-        candidates = cls.find_by_predicate(module, version, predicate, **kwargs)
-        if candidates:
-            return candidates[-1]
-
-        return cls.make(module, version, **kwargs)
-
-    def on_compute_predicate(self) -> Optional[Dict]:
-        """Event to compute a custom predicate dictionary of the element
-        where each key presents a predicate that can be used
-        during element lookup. Keys marked with a * are automatically
-        matched."""
+        return predicate
 
     @property
     def predicate(self) -> DictConfig:
         if self._predicate is None:
-            if self.__model__.predicate is None:
-                self.__model__.predicate = OmegaConf.to_container(
-                    OmegaConf.create(self.compute_predicate())
-                )
-
+            if not self.__model__.predicate:
+                return OmegaConf.create({})
             self._predicate = OmegaConf.create(self.__model__.predicate)
 
         return self._predicate
@@ -548,11 +478,10 @@ class Element(Mixin, Jsonable):
                         replace=True,
                     )
 
-                    # expose raw config so events and config methods can use it
-                    config_model = from_element(self) or pydantic.BaseModel
-                    raw_config = config_model().dict()
+                    # expose default config so events and config methods can use it
+                    default_config, config_model = from_element(self)
 
-                    self._config = OmegaConf.create(raw_config)
+                    self._config = OmegaConf.create(default_config)
 
                     self.on_before_configure(self._config)
 
@@ -601,10 +530,11 @@ class Element(Mixin, Jsonable):
                     config = OmegaConf.to_container(self._config, resolve=True)
 
                     # enforce schema
-                    config = config_model(**config).dict()
+                    if config_model is not None:
+                        config = config_model(**config).dict()
 
                     # add introspection data
-                    config["_raw_"] = raw_config
+                    config["_default_"] = default_config
                     config["_version_"] = __version
                     config["_update_"] = config_update
 
@@ -629,35 +559,9 @@ class Element(Mixin, Jsonable):
     def lineage(self) -> Tuple[str, ...]:
         return self.__model__.lineage
 
-    @property
-    def storage_id(self) -> Optional[str]:
-        if not self.is_mounted():
-            return None
-
-        return self.__model__._storage_id
-
-    @property
-    def storage_instance(self) -> Optional["Storage"]:
-        if not self.is_mounted():
-            return None
-
-        return self.__model__._storage_instance
-
     @classmethod
-    def from_storage(cls, storage_id, storage=None) -> "Element":
-        if storage is None:
-            from machinable.storage import Storage
-
-            storage = Storage.get()
-
-        return cls.from_model(
-            getattr(storage, f"retrieve_{cls.kind.lower()}")(storage_id)
-        )
-
-    @classmethod
-    def collect(cls, elements) -> Collection:
-        """Returns a collection of the element type"""
-        return Collection(elements)
+    def collect(cls, elements) -> ElementCollection:
+        return ElementCollection(elements)
 
     @classmethod
     def model(cls, element: Optional[Any] = None) -> schema.Element:
@@ -675,14 +579,41 @@ class Element(Mixin, Jsonable):
 
         return getattr(schema, cls.kind)
 
-    def matches(self, element: "Element", predicate: str) -> bool:
-        for p in resolve_custom_predicate(predicate, element):
-            if not equalversion(self.predicate[p], element.predicate[p]):
-                return False
+    def matches(
+        self, element: "Element", context: Optional[Dict] = sentinel
+    ) -> bool:
+        if context is sentinel:
+            context = element.compute_context()
+
+        if context is None:
+            # full constraint, match none
+            return False
+
+        if not context:
+            # no constraints, match all
+            return True
+
+        for field, value in context.items():
+            if field in (
+                "uuid",
+                "kind",
+                "module",
+            ):
+                if not equaljson(getattr(self, field), value):
+                    return False
+            elif field == "config":
+                if not equaljson(self._enderscore_config, value):
+                    return False
+            elif field == "predicate":
+                for p, v in value.items():
+                    if not equaljson(self.predicate[p], v):
+                        return False
+            else:
+                raise ValueError("Invalid context field: {field}")
 
         return True
 
-    def set_model(self, model: schema.Element) -> "Element":
+    def set_model(self, model: schema.Element) -> Self:
         self.__model__ = model
 
         # invalidate cached config
@@ -691,54 +622,6 @@ class Element(Mixin, Jsonable):
 
         return self
 
-    def local_directory(
-        self, *append: str, create: bool = False
-    ) -> Optional[str]:
-        if not self.is_mounted():
-            return None
-
-        return self.__model__._storage_instance.local_directory(
-            self, *append, create=create
-        )
-
-    def load_file(self, filepath: str, default=None) -> Optional[Any]:
-        if not self.is_mounted():
-            # has write been deferred?
-            if filepath in self._deferred_data:
-                return self._deferred_data[filepath]
-
-            return default
-
-        data = self.__model__._storage_instance.retrieve_file(self, filepath)
-
-        return data if data is not None else default
-
-    def save_file(self, filepath: str, data: Any) -> str:
-        if os.path.isabs(filepath):
-            raise ValueError("Filepath must be relative")
-
-        if not self.is_mounted():
-            # defer writes until element storage creation
-            self._deferred_data[filepath] = data
-            return "$deferred"
-
-        file = self.__model__._storage_instance.create_file(
-            self, filepath, data
-        )
-
-        # mark scripts as executable
-        if filepath.endswith(".sh"):
-            st = os.stat(file)
-            os.chmod(file, st.st_mode | stat.S_IEXEC)
-
-        return file
-
-    def save_data(self, filepath: str, data: Any) -> str:
-        return self.save_file(os.path.join("data", filepath), data)
-
-    def load_data(self, filepath: str, default=None) -> Optional[Any]:
-        return self.load_file(os.path.join("data", filepath), default)
-
     def serialize(self) -> Dict:
         # ensure that configuration has been parsed
         assert self.config is not None
@@ -752,67 +635,6 @@ class Element(Mixin, Jsonable):
     def is_connected(cls) -> bool:
         return len(_CONNECTIONS[cls.kind]) > 0
 
-    def to_cli(self) -> str:
-        cli = [self.module]
-        for v in self.__model__.version:
-            if isinstance(v, str):
-                cli.append(v)
-            else:
-                cli.extend(
-                    [
-                        f"{key}={shlex.quote(str(val))}"
-                        for key, val in flatten(v, reducer="dot").items()
-                    ]
-                )
-
-        return " ".join(cli)
-
-    def __enter__(self):
-        _CONNECTIONS[self.kind].append(self)
-        return self
-
-    def __exit__(self, *args, **kwargs):
-        try:
-            _CONNECTIONS[self.kind].pop()
-        except IndexError:
-            pass
-
-    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
-        return (self.__class__, (), self.serialize())
-
-    def __getstate__(self):
-        return self.serialize()
-
-    def __setstate__(self, state):
-        self.__model__ = self.__class__.model()(**state)
-
-    def __str__(self):
-        return self.__repr__()
-
-    def compute_predicate(self) -> Dict:
-        from machinable.project import Project
-
-        custom = self.on_compute_predicate() or {}
-        if isinstance(self, Project):
-            custom.update(self.global_predicate() or {})
-        else:
-            custom.update(Project.get().provider().global_predicate() or {})
-
-        config = copy.deepcopy(OmegaConf.to_container(self.config))
-        raw = config.pop("_raw_")
-        version = config.pop("_version_")
-        update = config.pop("_update_")
-
-        return {
-            "config_update": _filter_enderscores(update),
-            "config_update_": update,
-            "config": _filter_enderscores(config),
-            "config_": config,
-            "version": idversion(version),
-            "version_": version,
-            **{k.replace("*", ""): v for k, v in custom.items()},
-        }
-
     def on_instantiate(self) -> None:
         """Event that is invoked whenever the element is instantiated"""
 
@@ -830,6 +652,10 @@ class Element(Mixin, Jsonable):
         that are applied at a later stage.
         """
 
+    def _clear_caches(self) -> None:
+        self._config = None
+        self.__model__.config = None
+
     def __getattr__(self, name) -> Any:
         attr = getattr(self.__mixin__, name, None)
         if attr is not None:
@@ -840,9 +666,49 @@ class Element(Mixin, Jsonable):
             )
         )
 
+    def __enter__(self) -> Self:
+        _CONNECTIONS[self.kind].append(self)
+        return self
+
+    def __exit__(self, *args, **kwargs):
+        try:
+            _CONNECTIONS[self.kind].pop()
+        except IndexError:
+            pass
+
+    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
+        return (self.__class__, (), self.serialize())
+
+    def __getstate__(self):
+        return self.serialize()
+
+    def __setstate__(self, state):
+        self.__model__ = self.__class__.model()(**state)
+
+    def __repr__(self):
+        return f"{self.kind} [{self.id}]"
+
+    def __str__(self):
+        return self.id
+
+    def __eq__(self, other):
+        return self.uuid == getattr(other, "uuid", None)
+
+    def __ne__(self, other):
+        return self.uuid != getattr(other, "uuid", None)
+
 
 def get_lineage(element: "Element") -> Tuple[str, ...]:
     return tuple(
         obj.module if isinstance(obj, Element) else obj.__module__
         for obj in element.__class__.__mro__[1:-3]
     )
+
+
+def get_dump(element: "Element") -> Optional[bytes]:
+    if element.__model__.module.startswith("__session__"):
+        return pickle.dumps(element.__class__)
+
+
+def reset_connections() -> None:
+    _CONNECTIONS.clear()
