@@ -4,21 +4,19 @@ import json
 import os
 
 from machinable.errors import ConfigurationError
-from machinable.schema import Execution, Experiment
 from machinable.types import ElementType
-from pydantic import BaseModel
+from pydantic import BaseModel, Field
 
 
 class Settings(BaseModel):
-    default_predicate: Optional[str] = "config,*"
     default_execution: Optional[ElementType] = None
-    default_experiment: Optional[ElementType] = None
+    default_component: Optional[ElementType] = None
+    default_interface: Optional[ElementType] = None
     default_schedule: Optional[ElementType] = None
     default_group: Optional[str] = "%Y_%U_%a/"
-    default_storage: List[Union[str, dict]] = [
-        "machinable.storage.filesystem",
-        {"directory": "./storage"},
-    ]
+    default_storage: Optional[ElementType] = None
+    default_index: Optional[ElementType] = None
+    default_scope: Optional[ElementType] = None
 
 
 def get_settings(file="~/.machinable/settings.json"):
