@@ -2,23 +2,30 @@ import types
 from types import ModuleType
 from typing import Any, Callable, List, Mapping, Optional, Tuple, Union
 
+import sys
+
+if sys.version_info >= (3, 11):
+    from typing import Self
+else:
+    from typing_extensions import Self
+
 import importlib
 import inspect
 import json
 import os
-import pickle
 import random
 import re
 import string
+import subprocess
 import sys
 from keyword import iskeyword
 from pathlib import Path
+from uuid import UUID
 
 import arrow
-import commandlib
+import dill as pickle
 import jsonlines
 import omegaconf
-from baseconv import base62
 from flatten_dict import unflatten as _unflatten_dict
 
 if sys.version_info >= (3, 8):
@@ -31,96 +38,17 @@ from omegaconf.omegaconf import OmegaConf
 sentinel = object()
 
 import json
-import threading
 
-from machinable.types import DatetimeType
-from observable import Observable
-
-
-class Events(Observable):
-    def __init__(self) -> None:
-        super().__init__()
-        self._heartbeat = None
-
-    def trigger(self, event: str, *args: Any, **kw: Any) -> list:
-        """Triggers all handlers which are subscribed to an event.
-        Returns True when there were callbacks to execute, False otherwise."""
-        # upstream pending on issue https://github.com/timofurrer/observable/issues/17
-        callbacks = list(self._events.get(event, []))
-        return [callback(*args, **kw) for callback in callbacks]
-
-    def heartbeats(self, seconds=10):
-        if self._heartbeat is not None:
-            self._heartbeat.cancel()
-
-        if seconds is None or int(seconds) == 0:
-            # disable heartbeats
-            return
-
-        def heartbeat():
-            t = threading.Timer(seconds, heartbeat)
-            t.daemon = True
-            t.start()
-            self.trigger("heartbeat")
-            return t
-
-        self._heartbeat = heartbeat()
-
-
-class Jsonable:
-    def as_json(self, stringify=True, **dumps_kwargs):
-        serialized = self.serialize()
-        if stringify:
-            serialized = json.dumps(serialized, **dumps_kwargs)
-        return serialized
-
-    @classmethod
-    def from_json(cls, serialized, **loads_kwargs):
-        if isinstance(serialized, str):
-            serialized = json.loads(serialized, **loads_kwargs)
-        return cls.unserialize(serialized)
-
-    def clone(self):
-        return self.__class__.from_json(self.as_json())
-
-    def serialize(self) -> dict:
-        raise NotImplementedError
-
-    @classmethod
-    def unserialize(cls, serialized: dict) -> Any:
-        raise NotImplementedError
-
-
-class Connectable:
-    """Connectable trait"""
-
-    __connection__: Optional["Connectable"] = None
-
-    @classmethod
-    def is_connected(cls) -> bool:
-        return cls.__connection__ is not None
-
-    @classmethod
-    def get(cls) -> "Connectable":
-        return cls() if cls.__connection__ is None else cls.__connection__
-
-    def __enter__(self):
-        self._outer_connection = (  # pylint: disable=attribute-defined-outside-init
-            self.__class__.__connection__
-        )
-        self.__class__.__connection__ = self
-        return self
-
-    def __exit__(self, *args, **kwargs):
-        if self.__class__.__connection__ is self:
-            self.__class__.__connection__ = None
-        if getattr(self, "_outer_connection", None) is not None:
-            self.__class__.__connection__ = self._outer_connection
-        return self
+from machinable.types import DatetimeType, VersionType
+from pydantic import BaseModel
 
 
 def serialize(obj):
     """JSON serializer for objects not serializable by default json code"""
+    if isinstance(obj, BaseModel):
+        return obj.dict()
+    if isinstance(obj, UUID):
+        return obj.hex
     if isinstance(obj, DatetimeType):
         return str(obj)
     if isinstance(obj, (omegaconf.DictConfig, omegaconf.ListConfig)):
@@ -144,70 +72,12 @@ def generate_seed(random_state=None):
     return random_state.randint(0, 2**31 - 1)
 
 
-def as_color(experiment_id: str):
-    return "".join(
-        encode_experiment_id(decode_experiment_id(i) % 16)
-        for i in experiment_id
-    )
-
-
 def timestamp_to_directory(timestamp: float) -> str:
     return (
         arrow.get(timestamp).strftime("%Y-%m-%dT%H%M%S_%f%z").replace("+", "_")
     )
 
 
-def encode_experiment_id(seed, or_fail=True) -> Optional[str]:
-    """Encodes a seed and returns the corresponding experiment ID
-
-    # Arguments
-    seed: int in the range 62^5 <= seed <= 62^6-1
-    or_fail: If True, raises a Value error instead of returning None
-    """
-    try:
-        if not isinstance(seed, int):
-            raise ValueError
-        if 62**5 <= seed <= 62**6 - 1:
-            return base62.encode(seed)
-        raise ValueError
-    except (ValueError, TypeError) as e:
-        if or_fail:
-            raise ValueError(
-                "Seed has to lie in range 62^5 <= seed <= 62^6-1"
-            ) from e
-        return None
-
-
-def decode_experiment_id(experiment_id, or_fail=True) -> Optional[int]:
-    """Decodes a experiment ID into the corresponding seed
-
-    # Arguments
-    experiment_id: The base62 experiment ID
-    or_fail: If True, raises a Value error instead of returning None
-    """
-    try:
-        if not isinstance(experiment_id, str):
-            raise ValueError
-        value = int(base62.decode(experiment_id))
-        if 62**5 <= value <= 62**6 - 1:
-            return value
-        raise ValueError
-    except (ValueError, TypeError) as e:
-        if or_fail:
-            raise ValueError(
-                f"'{experiment_id}' is not a valid experiment ID"
-            ) from e
-        return None
-
-
-def generate_experiment_id(random_state=None) -> int:
-    if random_state is None or isinstance(random_state, int):
-        random_state = random.Random(random_state)
-
-    # ~ 55x10^9 distinct experiment IDs that if represented in base62 are len 6
-    return random_state.randint(62**5, 62**6 - 1)
-
-
 def is_valid_variable_name(name):
     if not isinstance(name, str):
         return False
@@ -218,7 +88,25 @@ def is_valid_module_path(name):
     return all(map(is_valid_variable_name, name.split(".")))
 
 
-def random_str(length, random_state=None):
+def is_directory_version(version: VersionType) -> bool:
+    if not isinstance(version, str):
+        return False
+    if not version.startswith("~"):
+        return True
+    if "/" in version or "\\" in version or version == "~":
+        return True
+
+    return False
+
+
+def joinpath(filepath: Union[str, List[str]]) -> str:
+    if isinstance(filepath, str):
+        return filepath
+
+    return os.path.join(*[str(p) for p in filepath if p is not None])
+
+
+def random_str(length: int, random_state=None):
     if random_state is None or isinstance(random_state, int):
         random_state = random.Random(random_state)
 
@@ -332,7 +220,7 @@ def generate_nickname(categories=None, glue="_"):
 
 
 def load_file(
-    filepath: str,
+    filepath: Union[str, List[str]],
     default: Any = sentinel,
     opener=open,
     **opener_kwargs,
@@ -347,6 +235,7 @@ def load_file(
     opener: Customer file opener
     opener_kwargs: Optional arguments to pass to the opener
     """
+    filepath = joinpath(filepath)
     _, ext = os.path.splitext(filepath)
     mode = opener_kwargs.pop("mode", "r")
     try:
@@ -381,7 +270,7 @@ def load_file(
 
 
 def save_file(
-    filepath: str,
+    filepath: Union[str, List[str]],
     data: Any,
     makedirs: Union[bool, Callable] = True,
     opener=open,
@@ -400,6 +289,7 @@ def save_file(
 
     Returns the absolute path to the written file
     """
+    filepath = joinpath(filepath)
     path = os.path.dirname(filepath)
     name = os.path.basename(filepath)
     _, ext = os.path.splitext(name)
@@ -588,59 +478,108 @@ def unflatten_dict(
 
 
 def get_diff(repository: str) -> Optional[str]:
-    git = commandlib.Command("git").in_dir(os.path.abspath(repository))
-
+    command = ["git", "diff", "--staged"]
     try:
-        return git("diff", "--staged").output()
-    except commandlib.exceptions.CommandError:
+        process = subprocess.run(
+            command,
+            cwd=os.path.abspath(repository),
+            capture_output=True,
+            text=True,
+            check=True,
+        )
+        return process.stdout
+    except subprocess.CalledProcessError:
         return None
 
 
-# This following method is modified 3rd party source code from
-# https://github.com/IDSIA/sacred/blob/7897c664b1b93fa2e2b6f3af244dfee590b1342a/sacred/dependencies.py#L401.
-# The copyright and license agreement can be found in the ThirdPartyNotices.txt file at the root of this repository.
-
-
 def get_commit(repository: str) -> dict:
     try:
-        import git
 
-        try:
-            repo = git.Repo(repository, search_parent_directories=False)
-            try:
-                branch = str(repo.active_branch)
-            except TypeError:
-                branch = None
+        def run_git_command(command):
+            return (
+                subprocess.check_output(
+                    ["git", "-C", repository] + command,
+                    stderr=subprocess.DEVNULL,
+                )
+                .decode()
+                .strip()
+            )
 
-            try:
-                path = repo.remote().url
-            except ValueError:
-                path = "git:/" + repo.working_dir
-            is_dirty = repo.is_dirty()
-            commit = repo.head.commit.hexsha
-            return {
-                "path": path,
-                "commit": commit,
-                "is_dirty": is_dirty,
-                "branch": branch,
-            }
-        except (git.exc.GitError, ValueError):
-            pass
-    except ImportError:
+        branch = run_git_command(["rev-parse", "--abbrev-ref", "HEAD"])
+        path = run_git_command(["config", "--get", "remote.origin.url"])
+        is_dirty = run_git_command(["status", "--porcelain"]) != ""
+        commit = run_git_command(["rev-parse", "HEAD"])
+
+        return {
+            "path": path,
+            "commit": commit,
+            "is_dirty": is_dirty,
+            "branch": branch,
+        }
+    except subprocess.CalledProcessError:
         pass
 
-    # todo: fallback
-
     return {"path": None, "commit": None, "is_dirty": None, "branch": None}
 
 
 def get_root_commit(repository: str) -> Optional[str]:
+    command = ["git", "rev-list", "--parents", "HEAD"]
     try:
-        return (
-            commandlib.Command("git", "rev-list", "--parents", "HEAD")
-            .in_dir(repository)
-            .output()[-1]
-            .replace("\n", "")
+        process = subprocess.run(
+            command, cwd=repository, capture_output=True, text=True, check=True
         )
-    except commandlib.exceptions.CommandError:
+        return process.stdout.split("\n")[-2]
+    except subprocess.CalledProcessError:
         return None
+
+
+class Jsonable:
+    def as_json(self, stringify=True, default=serialize, **dumps_kwargs):
+        serialized = self.serialize()
+        if stringify:
+            serialized = json.dumps(serialized, default=default, **dumps_kwargs)
+        return serialized
+
+    @classmethod
+    def from_json(cls, serialized, **loads_kwargs):
+        if isinstance(serialized, str):
+            serialized = json.loads(serialized, **loads_kwargs)
+        return cls.unserialize(serialized)
+
+    def clone(self):
+        return self.__class__.from_json(self.as_json())
+
+    def serialize(self) -> dict:
+        raise NotImplementedError
+
+    @classmethod
+    def unserialize(cls, serialized: dict) -> Any:
+        raise NotImplementedError
+
+
+class Connectable:
+    """Connectable trait"""
+
+    __connection__: Optional["Connectable"] = None
+
+    @classmethod
+    def is_connected(cls) -> bool:
+        return cls.__connection__ is not None
+
+    @classmethod
+    def get(cls) -> "Connectable":
+        return cls() if cls.__connection__ is None else cls.__connection__
+
+    def __enter__(self) -> Self:
+        self._outer_connection = (  # pylint: disable=attribute-defined-outside-init
+            self.__class__.__connection__
+        )
+        self.__class__.__connection__ = self
+        return self
+
+    def __exit__(self, *args, **kwargs) -> Self:
+        if self.__class__.__connection__ is self:
+            self.__class__.__connection__ = None
+        if getattr(self, "_outer_connection", None) is not None:
+            self.__class__.__connection__ = self._outer_connection
+        return self
